/* Functions for creating and destroying the log window, ouputting strings,
 * processing messages and so on.
 *
 * Written and copyright 1999 Adam Lock <locka@iol.ie>
 *
 * Distributed under the GNU General Public License; see the README file.
 * This code comes with NO WARRANTY.
 *
 * $Id: w32log.c,v 1.1 2000/09/16 15:08:50 swa Exp $
 *
 */
#include <assert.h>
#include <stdio.h>

#include <windows.h>
#include <richedit.h>

#include "gnu_regex.h"
#include "jcc.h"
#include "w32log.h"
#include "w32taskbar.h"
#include "w32rulesdlg.h"

#include "w32res.h"
#include <afxres.h>

extern char *win32_blurb;
extern HINSTANCE g_hInstance;
extern int g_nCmdShow;

/*
 * Timers and the various durations
 */
#define TIMER_ANIM_ID               1
#define TIMER_ANIM_TIME             100
#define TIMER_ANIMSTOP_ID           2
#define TIMER_ANIMSTOP_TIME         1000
#define TIMER_CLIPBUFFER_ID         3
#define TIMER_CLIPBUFFER_TIME       1000
#define TIMER_CLIPBUFFER_FORCE_ID   4
#define TIMER_CLIPBUFFER_FORCE_TIME 5000

/*
 * Styles of text that can be output
 */
#define STYLE_NONE      0
#define STYLE_HIGHLIGHT 1
#define STYLE_LINK      2
#define STYLE_HEADER    3

/*
 * Number of frames of animation in tray activity sequence
 */
#define ANIM_FRAMES 8

/*
 * Stored settings and their default values
 */
#define SETTING_FILE ".\\ijbw32.ini"
#define SETTING_SHOW_ACTIVITY_ANIMATION "activity-animation"
#define SETTING_SHOW_ON_TASKBAR         "show-on-task-bar"
#define SETTING_MINIMIZE_ON_CLOSE       "close-button-minimizes"
#define SETTING_LOG_MESSAGES            "log-messages"
#define SETTING_HIGHLIGHT_MESSAGES      "log-highlight-messages"
#define SETTING_LIMIT_BUFFER_SIZE       "log-buffer-size"
#define SETTING_MAX_BUFFER_LINES        "log-max-lines"
#define SETTING_LOG_FONT_NAME           "log-font-name"
#define SETTING_LOG_FONT_SIZE           "log-font-size"

#define DEFAULT_SHOW_ACTIVITY_ANIMATION TRUE
#define DEFAULT_SHOW_ON_TASKBAR         FALSE
#define DEFAULT_MINIMIZE_ON_CLOSE       TRUE
#define DEFAULT_LOG_MESSAGES            TRUE
#define DEFAULT_HIGHLIGHT_MESSAGES      TRUE
#define DEFAULT_LIMIT_BUFFER_SIZE       TRUE
#define DEFAULT_MAX_BUFFER_LINES        200
#define DEFAULT_LOG_FONT_NAME           "MS Sans Serif"
#define DEFAULT_LOG_FONT_SIZE           8

/*
 * These values affect the way the log window behaves, they should be read
 * from a file but for the moment, they are hardcoded here. Some options are
 * configurable through the UI.
 */
static BOOL g_bShowActivityAnimation = DEFAULT_SHOW_ACTIVITY_ANIMATION; /* Indicates whether task bar shows activity animation */
static BOOL g_bShowOnTaskBar = DEFAULT_SHOW_ON_TASKBAR; /* Indicates if the log window appears on the task bar */
static BOOL g_bCloseHidesWindow = DEFAULT_MINIMIZE_ON_CLOSE; /* Indicates whether closing the log window really just hides it */
static BOOL g_bLogMessages = DEFAULT_LOG_MESSAGES; /* Indicates if messages are logged at all */
static BOOL g_bHighlightMessages = DEFAULT_HIGHLIGHT_MESSAGES; /* Indicates whether log messages are highlighted */
static BOOL g_bLimitBufferSize = DEFAULT_LIMIT_BUFFER_SIZE; /* Indicates if buffer is limited in size */
static int g_nMaxBufferLines = DEFAULT_MAX_BUFFER_LINES; /* Maximum number of lines allowed in buffer when limited */
static char g_szFontFaceName[255] = DEFAULT_LOG_FONT_NAME; /* Font to use */
static int g_nFontSize = DEFAULT_LOG_FONT_SIZE; /* Size of font to use */

/* Regular expression for detected URLs */
#define RE_URL "http:[^ \n\r]*"

/*
 * Regular expressions that are used to perform highlight in the log window
 */
static struct _Pattern
{
	const char *str;
	int style;
	regex_t buffer;
} patterns_to_highlight[] =
{
	/* url headers */
	{ RE_URL,                STYLE_LINK },
/*	{ "[a-zA-Z0-9]+\\.[a-zA-Z0-9]+\\.[a-zA-Z0-9]+\\.[^ \n\r]*", STYLE_LINK }, */
	/* interesting text to highlight */
	{ "crunch!",		     STYLE_HIGHLIGHT },
	/* http headers */
	{ "referer:",            STYLE_HEADER },
	{ "proxy-connection:",   STYLE_HEADER },
	{ "proxy-agent:",        STYLE_HEADER },
	{ "user-agent:",         STYLE_HEADER },
	{ "host:",               STYLE_HEADER },
	{ "accept:",             STYLE_HEADER },
	{ "accept-encoding:",    STYLE_HEADER },
	{ "accept-language:",    STYLE_HEADER },
	{ "accept-charset:",     STYLE_HEADER },
	{ "accept-ranges:",      STYLE_HEADER },
	{ "date:",               STYLE_HEADER },
	{ "cache-control:",      STYLE_HEADER },
    { "cache-last-checked:", STYLE_HEADER },
	{ "connection:",         STYLE_HEADER },
	{ "content-type",        STYLE_HEADER },
	{ "content-length",      STYLE_HEADER },
	{ "cookie",              STYLE_HEADER },
	{ "last-modified:",      STYLE_HEADER },
	{ "pragma:",             STYLE_HEADER },
	{ "server:",             STYLE_HEADER },
	{ "etag:",               STYLE_HEADER },
	{ "expires:",            STYLE_HEADER },
	{ "warning:",            STYLE_HEADER },
	/* this is the terminator statement - do not delete! */
	{ NULL,                  STYLE_NONE }
};


/*
 * Public variables
 */
HWND g_hwndLogFrame;

/*
 * Private variables
 */
static CRITICAL_SECTION g_criticalsection;
static HWND g_hwndTray;
static HWND g_hwndLogBox;
static WNDPROC g_fnLogBox;
static HICON g_hiconAnim[ANIM_FRAMES];
static HICON g_hiconIdle;
static HICON g_hiconApp;
static int g_nAnimFrame;
static BOOL g_bClipPending = FALSE;
static int g_nRichEditVersion = 0;

/*
 * Private functions
 */
static HWND CreateLogWindow(HINSTANCE hInstance, int nCmdShow);
static HWND CreateHiddenLogOwnerWindow(HINSTANCE hInstance);
static LRESULT CALLBACK LogWindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK LogOwnerWindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static LRESULT CALLBACK LogRichEditProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
static BOOL InitRichEdit();
static void LogClipBuffer();
static void LogCreatePatternMatchingBuffers();
static void LogDestroyPatternMatchingBuffers();
static int LogPutStringNoMatch(const char *pszText, int style);
static void LoadLogSettings();
static void SaveLogSettings();

/*
 * Initialise the log window
 */
BOOL InitLogWindow()
{
	int i;

	/* Initialise the settings */
	LoadLogSettings();

	/* Load the icons */
	g_hiconIdle = LoadIcon(g_hInstance, MAKEINTRESOURCE(IDI_IDLE));
	for (i = 0; i < ANIM_FRAMES; i++)
	{
		g_hiconAnim[i] = LoadIcon(g_hInstance, MAKEINTRESOURCE(IDI_JUNKBUSTER1 + i));
	}
	g_hiconApp = LoadIcon(g_hInstance, MAKEINTRESOURCE(IDI_JUNKBUSTER));

	/* Create the user interface */
	g_hwndLogFrame = CreateLogWindow(g_hInstance, g_nCmdShow);
	g_hwndTray = CreateTrayWindow(g_hInstance);
	TrayAddIcon(g_hwndTray, 1, g_hiconApp, "Junkbuster");

	/* Create pattern matching buffers (for highlighting */
	LogCreatePatternMatchingBuffers();

	/* Create a critical section to protect multi-threaded access to certain things */
	InitializeCriticalSection(&g_criticalsection); 

	return TRUE;
}


/*
 * Cleanup the logwindow
 */
void TermLogWindow()
{
	int i;

	LogDestroyPatternMatchingBuffers();

	TrayDeleteIcon(g_hwndTray, 1);
	DeleteObject(g_hiconApp);
	DeleteObject(g_hiconIdle);
	for (i = 0; i < ANIM_FRAMES; i++)
	{
		DeleteObject(g_hiconAnim[i]);
	}
}


/*
 * Loads an integer setting from the settings file
 */
UINT LoadLogIntSetting(const char *szSetting, INT nDefault)
{
	return GetPrivateProfileInt("IJB", szSetting, nDefault, SETTING_FILE);
}


/*
 * Loads a string setting from the settings file
 */
DWORD LoadLogStringSetting(const char *szSetting, const char *szDefault, char *szBuffer, int nBufferSize)
{
	return GetPrivateProfileString("IJB", szSetting, szDefault, szBuffer, nBufferSize, SETTING_FILE);
}


/*
 * Save a string setting to the setting file
 */
void SaveLogStringSetting(const char *szSetting, const char *szValue)
{
	WritePrivateProfileString("IJB", szSetting, szValue, SETTING_FILE);
}


/*
 * Save an integer setting to the setting file
 */
void SaveLogIntSetting(const char *szSetting, INT nValue)
{
	char szValue[100];
	sprintf(szValue, "%d", nValue);
	SaveLogStringSetting(szSetting, szValue);
}


/*
 * Loads the log window settings
 */
void LoadLogSettings()
{
	g_bShowActivityAnimation = LoadLogIntSetting(SETTING_SHOW_ACTIVITY_ANIMATION, g_bShowActivityAnimation);
	g_bShowOnTaskBar = LoadLogIntSetting(SETTING_SHOW_ON_TASKBAR, g_bShowOnTaskBar);
	g_bShowActivityAnimation = LoadLogIntSetting(SETTING_MINIMIZE_ON_CLOSE, g_bCloseHidesWindow);
	g_bLogMessages = LoadLogIntSetting(SETTING_LOG_MESSAGES, g_bLogMessages);
	g_bHighlightMessages = LoadLogIntSetting(SETTING_HIGHLIGHT_MESSAGES, g_bHighlightMessages);
	g_bLimitBufferSize = LoadLogIntSetting(SETTING_LIMIT_BUFFER_SIZE, g_bLimitBufferSize);
	g_nMaxBufferLines = LoadLogIntSetting(SETTING_MAX_BUFFER_LINES, g_nMaxBufferLines);
	memset(g_szFontFaceName, 0, sizeof(g_szFontFaceName));
	LoadLogStringSetting(SETTING_LOG_FONT_NAME, DEFAULT_LOG_FONT_NAME, g_szFontFaceName, sizeof(g_szFontFaceName) / sizeof(g_szFontFaceName[0]));
	g_nFontSize = LoadLogIntSetting(SETTING_LOG_FONT_SIZE, DEFAULT_LOG_FONT_SIZE);
}


/*
 * Saves the log window settings from their current values
 */
void SaveLogSettings()
{
	SaveLogIntSetting(SETTING_SHOW_ACTIVITY_ANIMATION, g_bShowActivityAnimation);
	SaveLogIntSetting(SETTING_SHOW_ON_TASKBAR, g_bShowOnTaskBar);
	SaveLogIntSetting(SETTING_MINIMIZE_ON_CLOSE, g_bCloseHidesWindow);
	SaveLogIntSetting(SETTING_LOG_MESSAGES, g_bLogMessages);
	SaveLogIntSetting(SETTING_HIGHLIGHT_MESSAGES, g_bHighlightMessages);
	SaveLogIntSetting(SETTING_LIMIT_BUFFER_SIZE, g_bLimitBufferSize);
	SaveLogIntSetting(SETTING_MAX_BUFFER_LINES, g_nMaxBufferLines);
	SaveLogStringSetting(SETTING_LOG_FONT_NAME, g_szFontFaceName);
	SaveLogIntSetting(SETTING_LOG_FONT_SIZE, g_nFontSize);
}


/*
 * Compile the pattern matching buffers
 */
void LogCreatePatternMatchingBuffers()
{
	int i;
	for (i = 0; patterns_to_highlight[i].str != NULL; i++)
	{
		regcomp(&patterns_to_highlight[i].buffer, patterns_to_highlight[i].str, REG_ICASE);
	}
}


/*
 * Free up the pattern matching buffers
 */
void LogDestroyPatternMatchingBuffers()
{
	int i;
	for (i = 0; patterns_to_highlight[i].str != NULL; i++)
	{
		regfree(&patterns_to_highlight[i].buffer);
	}
}

/*
 * Returns the URL from under the cursor (remember to free it!)
 */
char *LogGetURLUnderCursor()
{
	regex_t re;
	POINT ptCursor;
	POINTL ptl;
	DWORD nPos;
	DWORD nWordStart = 0;
	DWORD nWordEnd = 0;
	char *szResult = NULL;
	
	regcomp(&re, RE_URL, REG_ICASE);

	/* Get the position of the cursor over the text window */
	GetCursorPos(&ptCursor);
	ScreenToClient(g_hwndLogBox, &ptCursor);
	ptl.x = ptCursor.x;
	ptl.y = ptCursor.y;

	/* Search backwards and fowards to obtain the word that is highlighted */
	nPos = LOWORD(SendMessage(g_hwndLogBox, EM_CHARFROMPOS, 0, (LPARAM) &ptl));
	nWordStart = SendMessage(g_hwndLogBox, EM_FINDWORDBREAK, WB_LEFT, nPos);
	nWordEnd = SendMessage(g_hwndLogBox, EM_FINDWORDBREAK, WB_RIGHTBREAK, nPos);

	/* Compare the string to the pattern */
	if (nWordEnd > nWordStart)
	{
		TEXTRANGE range;
		regmatch_t match;

		range.chrg.cpMin = nWordStart;
		range.chrg.cpMax = nWordEnd;
		range.lpstrText = malloc(nWordEnd - nWordStart + 1);
		memset(range.lpstrText, 0, nWordEnd - nWordStart + 1);
		SendMessage(g_hwndLogBox, EM_GETTEXTRANGE, 0, (LPARAM) &range);

		if (regexec(&re, range.lpstrText, 1, &match, 0) == 0)
		{
			szResult = range.lpstrText;
		}
		else
		{
			free(range.lpstrText);
		}

		regfree(&re);
	}

	return szResult;
}


/*
 * Inserts text into the logging window
 */
int LogPutString(const char *pszText)
{
	int i;
	int result = 0;

	if (pszText == NULL || strlen(pszText) == 0)
	{
		return 1;
	}

	if (!g_bLogMessages)
	{
		return 1;
	}

	/* Critical section stops multiple threads doing nasty interactions that
	 * foul up the highlighting and output.
	 */
	EnterCriticalSection(&g_criticalsection); 

	if (g_bHighlightMessages)
	{
		regmatch_t match;

		/* First things first, regexp scan for various things that we would like highlighted */
		for (i = 0; patterns_to_highlight[i].str != NULL; i++)
		{
			if (regexec(&patterns_to_highlight[i].buffer, pszText, 1, &match, 0) == 0)
			{
				char *pszBefore = NULL;
				char *pszMatch = NULL;
				char *pszAfter = NULL;
				int nMatchSize;

				/* Split the string up into pieces representing the strings, before
				   at and after the matching pattern
				 */
				if (match.rm_so > 0)
				{
					pszBefore = malloc((match.rm_so + 1) * sizeof(char));
					memset(pszBefore, 0, (match.rm_so + 1) * sizeof(char));
					strncpy(pszBefore, pszText, match.rm_so);
				}
				if (match.rm_eo < strlen(pszText))
				{
					pszAfter = strdup(&pszText[match.rm_eo]);
				}
				nMatchSize = match.rm_eo - match.rm_so;
				pszMatch = malloc(nMatchSize + 1);
				strncpy(pszMatch, &pszText[match.rm_so], nMatchSize);
				pszMatch[nMatchSize] = '\0';

				/* Recursively call LogPutString */
				if (pszBefore)
				{
					LogPutString(pszBefore);
					free(pszBefore);
				}
				if (pszMatch)
				{
					LogPutStringNoMatch(pszMatch, patterns_to_highlight[i].style);
					free(pszMatch);
				}
				if (pszAfter)
				{
					LogPutString(pszAfter);
					free(pszAfter);
				}

				result = 1;
				goto end;
			}
		}
	}

	result = LogPutStringNoMatch(pszText, STYLE_NONE);

end:
	LeaveCriticalSection(&g_criticalsection);

	return result;
}


/*
 * Puts a string into the buffer
 */
int LogPutStringNoMatch(const char *pszText, int style)
{
	CHARRANGE range;
	CHARFORMAT format;
	int nTextLength;

	assert(g_hwndLogBox);
	if (g_hwndLogBox == NULL)
	{
		return 1;
	}

	/* TODO preserve existing selection */

	/* Go to the end of the text */
	nTextLength = GetWindowTextLength(g_hwndLogBox);
   	range.cpMin = nTextLength;
	range.cpMax = nTextLength;
	SendMessage(g_hwndLogBox, EM_EXSETSEL, 0, (LPARAM) &range);

	/* Apply a formatting style */
	memset(&format, 0, sizeof(format));
	format.cbSize = sizeof(format);
	format.dwMask = CFM_BOLD | CFM_UNDERLINE | CFM_STRIKEOUT | CFM_ITALIC | CFM_COLOR | CFM_FACE | CFM_SIZE;
	format.yHeight = (g_nFontSize * 1440) / 72;
	strcpy(format.szFaceName, g_szFontFaceName);
	if (style == STYLE_NONE)
	{
		/* DO NOTHING */
		format.dwEffects |= CFE_AUTOCOLOR;
	}
	else if (style == STYLE_HEADER)
	{
		format.dwEffects |= CFE_AUTOCOLOR | CFE_ITALIC;
	}
	else if (style == STYLE_HIGHLIGHT)
	{
		format.dwEffects |= CFE_AUTOCOLOR | CFE_BOLD;
	}
	else if (style == STYLE_LINK)
	{
		format.dwEffects |= CFE_UNDERLINE;
		format.crTextColor = RGB(0, 0, 255);
	}
	SendMessage(g_hwndLogBox, EM_SETCHARFORMAT, SCF_SELECTION, (LPARAM) &format);

	/* Append text to the end */
	SendMessage(g_hwndLogBox, EM_REPLACESEL, FALSE, (LPARAM) pszText);

	/* TODO Restore the old selection */

	/* Purge buffer */
	if (strchr(pszText, '\n') != NULL)
	{
		SetTimer(g_hwndLogFrame, TIMER_CLIPBUFFER_ID, TIMER_CLIPBUFFER_TIME, NULL);
		if (!g_bClipPending)
		{
			/* Set the force clip timer going. This timer ensures clipping is done
			   intermittently even when there is a sustained burst of logging
			*/
			SetTimer(g_hwndLogFrame, TIMER_CLIPBUFFER_FORCE_ID, TIMER_CLIPBUFFER_FORCE_TIME, NULL);
		}
		g_bClipPending = TRUE;
	}

	return 1;
}

/*
 * Start the spinner
 */
void LogShowActivity()
{
	/* Start some activity timers */
	if (g_bShowActivityAnimation)
	{
		SetTimer(g_hwndLogFrame, TIMER_ANIM_ID, TIMER_ANIM_TIME, NULL);
		SetTimer(g_hwndLogFrame, TIMER_ANIMSTOP_ID, TIMER_ANIMSTOP_TIME, NULL);
	}
}

/*
 * Prunes old lines from the log
 */
void LogClipBuffer()
{
	int nLines = SendMessage(g_hwndLogBox, EM_GETLINECOUNT, 0, 0);
	if (g_bLimitBufferSize && nLines > g_nMaxBufferLines)
	{
		/* Compute the range representing the lines to be deleted */
		LONG nLastLineToDelete = nLines - g_nMaxBufferLines;
		LONG nLastChar = SendMessage(g_hwndLogBox, EM_LINEINDEX, nLastLineToDelete, 0);
		CHARRANGE range;
		range.cpMin = 0;
		range.cpMax = nLastChar;

		/* TODO get current selection */

		/* TODO adjust and clip old selection against range to be deleted */

		/* Select range and erase it (turning off autoscroll to prevent
		   nasty scrolling) */
		SendMessage(g_hwndLogBox, EM_SETOPTIONS, ECOOP_XOR, ECO_AUTOVSCROLL);
		SendMessage(g_hwndLogBox, EM_EXSETSEL, 0, (LPARAM) &range);
		SendMessage(g_hwndLogBox, EM_REPLACESEL, FALSE, (LPARAM) "");
		SendMessage(g_hwndLogBox, EM_SETOPTIONS, ECOOP_XOR, ECO_AUTOVSCROLL);

		/* Restore old selection */
	}
}


/*
 * Creates a hidden owner window that stops the log window appearing in the task bar
 */
HWND CreateHiddenLogOwnerWindow(HINSTANCE hInstance)
{
	static const char *szWndName = "JunkbusterLogLogOwner";
	WNDCLASS wc;
	HWND hwnd;

	wc.style			= 0;
	wc.lpfnWndProc		= LogOwnerWindowProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= hInstance;
	wc.hIcon			= 0;
	wc.hCursor			= 0;
	wc.hbrBackground	= 0;
	wc.lpszMenuName		= 0;
	wc.lpszClassName	= szWndName;

	RegisterClass(&wc);

	hwnd = CreateWindow(szWndName, szWndName,
		WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		CW_USEDEFAULT, NULL, NULL, hInstance, NULL );

	return hwnd;
}


/*
 * Dummy procedure that does nothing special
 */
LRESULT CALLBACK LogOwnerWindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}


/*
 * Create the logging window
 */
HWND CreateLogWindow(HINSTANCE hInstance, int nCmdShow)
{
	static const char *szWndName = "JunkbusterLogWindow";
	static const char *szWndTitle = "Junkbuster";

	HWND hwnd = NULL;
	HWND hwndOwner = (g_bShowOnTaskBar) ? NULL : CreateHiddenLogOwnerWindow(hInstance);
	HWND hwndChild = NULL;
	RECT rcClient;
	WNDCLASSEX wc;

	memset(&wc, 0, sizeof(wc));
	wc.cbSize           = sizeof(wc);
	wc.style			= CS_DBLCLKS;
	wc.lpfnWndProc		= LogWindowProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= hInstance;
	wc.hIcon			= g_hiconApp;
	wc.hCursor			= 0;
	wc.hbrBackground	= 0;
	wc.lpszMenuName		= MAKEINTRESOURCE(IDR_LOGVIEW);
	wc.lpszClassName	= szWndName;
	wc.hbrBackground    = GetStockObject(WHITE_BRUSH); 
	RegisterClassEx(&wc);

	hwnd = CreateWindowEx(0, szWndName, szWndTitle,
		WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		CW_USEDEFAULT, hwndOwner, NULL, hInstance, NULL);

	/* Now create a child list box */
	GetClientRect(hwnd, &rcClient);

	/* Create a rich edit control */
	InitRichEdit();
	g_hwndLogBox = CreateWindowEx(0, (g_nRichEditVersion == 0x0100) ? "RichEdit" : RICHEDIT_CLASS, "",
		ES_AUTOVSCROLL | ES_MULTILINE | ES_READONLY | ES_NOHIDESEL | WS_CHILD | WS_VSCROLL | WS_HSCROLL | WS_VISIBLE,
		rcClient.left, rcClient.top, rcClient.right,	rcClient.bottom,
		hwnd, NULL, hInstance, NULL);
/*	SendMessage(g_hwndLogBox, EM_SETWORDWRAPMODE, 0, 0); */

	/* Subclass the control to catch certain messages */
	g_fnLogBox = (WNDPROC) GetWindowLong(g_hwndLogBox, GWL_WNDPROC);
	SetWindowLong(g_hwndLogBox, GWL_WNDPROC, (LONG) LogRichEditProc);

	/* Minimizing looks stupid when the log window is not on the task bar, so hide instead */
	if (!g_bShowOnTaskBar &&
			(nCmdShow == SW_SHOWMINIMIZED ||
			 nCmdShow == SW_MINIMIZE ||
			 nCmdShow == SW_SHOWMINNOACTIVE))
	{
		nCmdShow = SW_HIDE;
	}

	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);

	GetClientRect(g_hwndLogFrame, &rcClient);
	SetWindowPos(g_hwndLogBox, NULL, rcClient.left, rcClient.top, rcClient.right - rcClient.left, rcClient.bottom - rcClient.top, SWP_NOZORDER);

	return hwnd;
}


/*
 * Initialise the rich edit control library
 */
BOOL InitRichEdit()
{
	static HINSTANCE hInstRichEdit;
	if (hInstRichEdit == NULL)
	{
		g_nRichEditVersion = 0;
		hInstRichEdit = LoadLibraryA("RICHED20.DLL");
		if (hInstRichEdit)
		{
			g_nRichEditVersion = _RICHEDIT_VER;
		}
		else
		{
			hInstRichEdit = LoadLibraryA("RICHED32.DLL");
			if (hInstRichEdit)
			{
				g_nRichEditVersion = 0x0100;
			}
		}
	}
	return (hInstRichEdit != NULL) ? TRUE : FALSE;
}


/*
 * Shows or hides the log window
 */
void ShowLogWindow(BOOL bShow)
{
	int nAction;
	if (bShow)
	{
		SetWindowPos(g_hwndLogFrame, HWND_TOP, 0, 0, 0, 0, SWP_SHOWWINDOW | SWP_NOSIZE | SWP_NOMOVE);
	}
	else if (g_bShowOnTaskBar)
	{
		ShowWindow(g_hwndLogFrame, SW_MINIMIZE);
	}
	else
	{
		ShowWindow(g_hwndLogFrame, SW_HIDE);
	}
}


/*
 * Opens the specified setting file for editing
 */
void EditFile(const char *filename)
{
	if (filename)
	{
		ShellExecute(g_hwndLogFrame, "open", filename, NULL, NULL, SW_SHOWNORMAL);
	}
}


/*--------------------------------------------------------------------------*/
/* Windows message handlers                                                 */
/*--------------------------------------------------------------------------*/


/*
 * Handler for WM_RBUTTONUP messages
 */
void OnLogRButtonUp(int nModifier, int x, int y)
{
	HMENU hMenu = LoadMenu(g_hInstance, MAKEINTRESOURCE(IDR_POPUP_SELECTION));
	if (hMenu != NULL)
	{
		HMENU hMenuPopup = GetSubMenu(hMenu, 0);
		char *szURL;

		/* Check if there is a selection */
		CHARRANGE range;
		SendMessage(g_hwndLogBox, EM_EXGETSEL, 0, (LPARAM) &range);
		if (range.cpMin == range.cpMax)
		{
			EnableMenuItem(hMenuPopup, ID_EDIT_COPY, MF_BYCOMMAND | MF_GRAYED);
		}
		else
		{
			EnableMenuItem(hMenuPopup, ID_EDIT_COPY, MF_BYCOMMAND | MF_ENABLED);
		}

		/* Check if cursor is over a link */
		szURL = LogGetURLUnderCursor();
		if (szURL)
		{
			MENUITEMINFO item;
			TCHAR szMenuItemTemplate[1000];
			char *szMenuItem;

			memset(&item, 0, sizeof(item));
			item.cbSize = sizeof(item);
			item.fMask = MIIM_TYPE | MIIM_ID | MIIM_STATE;
			item.fType = MFT_STRING;
			item.fState = MFS_ENABLED;
			item.wID = ID_NEW_BLOCKER;

			/* Put the item into the menu */
			memset(szMenuItemTemplate, 0, sizeof(szMenuItemTemplate));
			LoadString(g_hInstance, IDS_NEW_BLOCKER, szMenuItemTemplate, sizeof(szMenuItemTemplate) / sizeof(szMenuItemTemplate[0]));

			szMenuItem = malloc(strlen(szMenuItemTemplate) + strlen(szURL) + 1);
			sprintf(szMenuItem, szMenuItemTemplate, szURL);

			item.dwTypeData = (DWORD) szMenuItem;
			item.cch = strlen(szMenuItem);

			InsertMenuItem(hMenuPopup, 1, TRUE, &item);

			SetDefaultRule(szURL);

			free(szURL);
		}

		/* Display the popup */
		TrackPopupMenu(hMenuPopup, TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RIGHTBUTTON, x, y, 0, g_hwndLogFrame, NULL);
		DestroyMenu(hMenu);
	}
}

/*
 * Handler for WM_COMMAND messages
 */
void OnLogCommand(int nCommand)
{
	switch (nCommand)
	{
	case ID_SHOWWINDOW:
		ShowLogWindow(TRUE);
		break;

	case ID_FILE_EXIT:
		PostMessage(g_hwndLogFrame, WM_CLOSE, 0, 0);
		break;
	
	case ID_EDIT_COPY:
		SendMessage(g_hwndLogBox, WM_COPY, 0, 0);
		break;

	case ID_VIEW_CLEARLOG:
		SendMessage(g_hwndLogBox, WM_SETTEXT, 0, (LPARAM) "");
		break;

	case ID_VIEW_LOGMESSAGES:
		g_bLogMessages = !g_bLogMessages;
		SaveLogSettings();
		break;

	case ID_VIEW_MESSAGEHIGHLIGHTING:
		g_bHighlightMessages = !g_bHighlightMessages;
		SaveLogSettings();
		break;

	case ID_VIEW_LIMITBUFFERSIZE:
		g_bLimitBufferSize = !g_bLimitBufferSize;
		SaveLogSettings();
		break;

	case ID_VIEW_ACTIVITYANIMATION:
		g_bShowActivityAnimation = !g_bShowActivityAnimation;
		SaveLogSettings();
		break;

	case ID_TOOLS_EDITJUNKBUSTER:
		EditFile("junkbstr.ini"); /* TODO get real name from somewhere */
		break;

	case ID_TOOLS_EDITCOOKIES:
		EditFile(cookiefile);
		break;

	case ID_TOOLS_EDITBLOCKERS:
		EditFile(blockfile);
		break;

	case ID_TOOLS_EDITFORWARD:
		EditFile(forwardfile);
		break;

	case ID_TOOLS_EDITTRUST:
		EditFile(trustfile);
		break;

	case ID_TOOLS_EDITIMAGE:
		EditFile(imagefile);
		break;

	case ID_TOOLS_EDITACLS:
		EditFile(aclfile);
		break;

	case ID_NEW_BLOCKER:
		ShowRulesDialog(g_hwndLogFrame);
		break;

	case ID_HELP_GPL:
		ShellExecute(g_hwndLogFrame, "open", "gpl.html", NULL, NULL, SW_SHOWNORMAL);
		break;

	case ID_HELP_FAQ:
		ShellExecute(g_hwndLogFrame, "open", "ijbfaq.html", NULL, NULL, SW_SHOWNORMAL);
		break;

	case ID_HELP_MANUAL:
		ShellExecute(g_hwndLogFrame, "open", "ijbman.html", NULL, NULL, SW_SHOWNORMAL);
		break;

	case ID_HELP_STATUS:
		ShellExecute(g_hwndLogFrame, "open", "Junkbuster Status.URL", NULL, NULL, SW_SHOWNORMAL);
		break;

	case ID_HELP_ABOUTJUNKBUSTER:
		MessageBox(g_hwndLogFrame, win32_blurb, "Junkbuster Information", MB_OK);
		break;

	default:
		/* DO NOTHING */
		break;
	}
}

/*
 * Handler for WM_INITMENU messages
 */
void OnLogInitMenu(HMENU hmenu)
{
	/* Only enable editors if there is a file to edit */
	EnableMenuItem(hmenu, ID_TOOLS_EDITCOOKIES, MF_BYCOMMAND | (cookiefile ? MF_ENABLED : MF_GRAYED));
	EnableMenuItem(hmenu, ID_TOOLS_EDITBLOCKERS, MF_BYCOMMAND | (blockfile ? MF_ENABLED : MF_GRAYED));
	EnableMenuItem(hmenu, ID_TOOLS_EDITFORWARD, MF_BYCOMMAND | (forwardfile ? MF_ENABLED : MF_GRAYED));
	EnableMenuItem(hmenu, ID_TOOLS_EDITTRUST, MF_BYCOMMAND | (trustfile ? MF_ENABLED : MF_GRAYED));
	EnableMenuItem(hmenu, ID_TOOLS_EDITIMAGE, MF_BYCOMMAND | (imagefile ? MF_ENABLED : MF_GRAYED));
	EnableMenuItem(hmenu, ID_TOOLS_EDITACLS, MF_BYCOMMAND | (aclfile ? MF_ENABLED : MF_GRAYED));

	/* Check/uncheck options */
	CheckMenuItem(hmenu, ID_VIEW_LOGMESSAGES, MF_BYCOMMAND | (g_bLogMessages ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hmenu, ID_VIEW_MESSAGEHIGHLIGHTING, MF_BYCOMMAND | (g_bHighlightMessages ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hmenu, ID_VIEW_LIMITBUFFERSIZE, MF_BYCOMMAND | (g_bLimitBufferSize ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem(hmenu, ID_VIEW_ACTIVITYANIMATION, MF_BYCOMMAND | (g_bShowActivityAnimation ? MF_CHECKED : MF_UNCHECKED));
}


/*
 * Handler for WM_TIMER messages
 */
void OnLogTimer(int nTimer)
{
	switch (nTimer)
	{
	
	case TIMER_ANIM_ID:
		TraySetIcon(g_hwndTray, 1, g_hiconAnim[g_nAnimFrame++ % ANIM_FRAMES]);
		break;
	
	case TIMER_ANIMSTOP_ID:
		g_nAnimFrame = 0;
		TraySetIcon(g_hwndTray, 1, g_hiconIdle);
		KillTimer(g_hwndLogFrame, TIMER_ANIM_ID);
		KillTimer(g_hwndLogFrame, TIMER_ANIMSTOP_ID);
		break;
	
	case TIMER_CLIPBUFFER_ID:
	case TIMER_CLIPBUFFER_FORCE_ID:
		LogClipBuffer();
		g_bClipPending = FALSE;
		KillTimer(g_hwndLogFrame, TIMER_CLIPBUFFER_ID);
		KillTimer(g_hwndLogFrame, TIMER_CLIPBUFFER_FORCE_ID);
		break;
	
	default:
		/* DO NOTHING */
		break;
	}
}

/*
 * Window subclass routine handles some events for the rich edit control
 */
LRESULT CALLBACK LogRichEditProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_RBUTTONUP:
		{
			POINT pt;
			pt.x = LOWORD(lParam);
			pt.y = HIWORD(lParam);
			ClientToScreen(hwnd, &pt);
			OnLogRButtonUp(wParam, pt.x, pt.y);
		}
		return 0;
	}
	return CallWindowProc(g_fnLogBox, hwnd, uMsg, wParam, lParam);
}


/*
 * Windows call back routine handles events on the log window
 */
LRESULT CALLBACK LogWindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_CREATE:
		return 0;
	
	case WM_CLOSE:
		/* This is the end - beautiful friend - the end */
		DestroyWindow(g_hwndLogBox);
		DestroyWindow(g_hwndLogFrame);
		return 0;

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;

	case WM_SHOWWINDOW:
	case WM_SIZE:
		/* Resize the logging window to fit the new frame */
		if (g_hwndLogBox)
		{
			RECT rc;
			GetClientRect(g_hwndLogFrame, &rc);
			SetWindowPos(g_hwndLogBox, NULL, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, SWP_NOZORDER);
		}
		return 0;

	case WM_INITMENU:
		OnLogInitMenu((HMENU) wParam);
		return 0;

	case WM_TIMER:
		OnLogTimer(wParam);
		return 0;

	case WM_COMMAND:
		OnLogCommand(LOWORD(wParam));
		return 0;

	case WM_SYSCOMMAND:
		switch (wParam)
		{
		case SC_CLOSE:
			if (g_bCloseHidesWindow)
			{
				ShowLogWindow(FALSE);
				return 0;
			}
			break;
		case SC_MINIMIZE:
			ShowLogWindow(FALSE);
			return 0;
		}
		break;
	}

	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}
 
